"""
    Copyright (c) 2021 David Pearson (david@seclarity.io)
    Date: 12/20/2021
    This file contains all of the main pieces needed to successfully stream unenriched secflows from a local system and
    upload them to NetworkSage.

    This software is provided under the Apache Software License.
    See the accompanying LICENSE file for more information.
"""

import argparse
import sys
import pickle
import sched
import threading
import time
import os
import platform
from pathlib import Path
from filelock import SoftFileLock, Timeout
from networksage_tools.streaming import captureutils
from networksage_tools.common_utilities import dnsservice
from networksage_tools.common_utilities import utilities

def schedule_cleanup(**kwargs):
    s = sched.scheduler(time.time, time.sleep)
    interval = kwargs['interval']

    def clean_up_short_term_passive_dns(scheduler):
        """The short-term passive DNS file is meant to keep track of DNS resolutions that may've happened recently,
            but that did not happen during a current streaming capture's time period. These entries should be expired
            regularly so as to not be over-zealous with labeling sessions. Since we have the long-term passive DNS file
            to capture DNS resolutions that don't change frequently, a good default time period to expire entries is
            five minutes (which is a commonly-used timeout for many popular sites).
        """
        short_term_passive_dns_filename = "./shortTermPassiveDNS.pkl"
        short_term_passive_dns_lock_name = "./shortTermPassiveDNS.pkl.lock"
        now = time.time()
        try:
            short_term_pdns_file = Path(short_term_passive_dns_filename)
            lock = SoftFileLock(short_term_passive_dns_lock_name)
            try:
                with lock.acquire(timeout=10):
                    cleaned_up_dict = dict()
                    with open(short_term_pdns_file, "rb") as stpdf:
                        short_term_pdns_dict = pickle.load(stpdf)
                        for entry in short_term_pdns_dict.keys():
                            for record in short_term_pdns_dict[entry]:
                                ts = record[0]
                            # use list comprehension to only keep those that aren't stale
                            cleaned_up_dict[entry] = [record for record in short_term_pdns_dict[entry] if (now - record[0] < 300)]
                            if len(cleaned_up_dict[entry]) == 0:
                                cleaned_up_dict.popitem()
                    with open(short_term_pdns_file, "wb") as stpdf:
                        pickle.dump(cleaned_up_dict, stpdf)
            except Timeout:
                print("Lock acquisition for cleaning up short-term passive DNS took too long. Something might be wrong.")
        except:
            pass # something went wrong, which may just be that there's no short term passive dns file
        s.enter(interval, 1, clean_up_short_term_passive_dns, (scheduler,))
    s.enter(interval, 1, clean_up_short_term_passive_dns, (s,))
    s.run()


def start(interface, duration=300, is_verbose=False):
    """Start the streaming functionality. Expects the name of the interface to capture network data from, and a duration
        (in seconds) to capture before uploading to NetworkSage. This function will run until killed, continually
        generating and uploading samples.
    """
    my_platform = platform.system().lower()
    if my_platform in ["linux", "darwin"]:
        os.nice(20) # Linux-specific, values are [-20,20] (higher being "nicer" to other processes)
    else:
        print("Need to implement resource limiting on Windows and other non-Linux systems (if it's too resource-intensive)!")
    api_key = os.environ.get("NETWORKSAGE_API_KEY")
    if api_key is None:
        print("Missing API Key. Please type export NETWORKSAGE_API_KEY='<your_api_key>' in your terminal to set up.")

    cleanup_thread = threading.Thread(target=schedule_cleanup, kwargs={"interval":60}) # clean up every minute
    cleanup_thread.start()

    start_time = time.time()
    while True:
        #set up threading
        utils = utilities.Utilities(None, my_platform)  # create an instance of utils to use
        capture_thread = threading.Thread(target=captureutils.capture
                                        , kwargs={"interface": interface
                                                , "bpf": captureutils.create_bpf()
                                                , "utils": utils
                                                , "is_verbose": is_verbose
                                                }
                                        )
        processing_thread = threading.Thread(target=captureutils.process_packets
                                            , kwargs={"utils":utils
                                                    , "is_verbose": is_verbose
                                                    }
                                            )
        iteration_timer = threading.Timer(duration
                                        , captureutils.send_sample
                                        , kwargs={"utils":utils
                                                , "capture_thread":capture_thread
                                                , "processing_thread":processing_thread
                                                , "key":None
                                                , "is_verbose": is_verbose
                                                }
                                        )
        print("Capturing on", interface, "for", str(duration), "seconds")

        # Start processing the packets we're collecting in our utilities packet buffer
        iteration_timer.start()
        capture_thread.start()
        processing_thread.start()
        time.sleep(duration - ((time.time() - start_time) % duration))

if __name__ == "__main__":

    # handle all arguments
    parser = argparse.ArgumentParser()
    streaming_group = parser.add_argument_group("Streaming Arguments", "arguments available when configuring streaming")
    streaming_group.add_argument("-i", "--interface", help="the name of the interface to capture network data from", type=str)
    streaming_group.add_argument("-d", "--duration", help="(optional) how long (in seconds) to capture before creating a new sample (default is 300 seconds)", type=int)

    args = parser.parse_args()
    if not args.interface:
        print("Error: Must specify an interface to capture on. Exiting!")
        sys.exit(1)
    if not args.duration:
        args.duration = 300
    if args.duration < 10:
        args.duration = 60
        print("Too low of a duration set. Setting duration to 60 seconds")
    start(args.interface, args.duration)


"""
    Copyright (c) 2021 David Pearson (david@seclarity.io)
    Date: 06/02/2021

    The functions in this file handle PCAP, PCPANG, or Zeek files. They remove all local-to-local traffic, then
     convert them into a Secflow format (which is similar to but more lightweight than Zeek flows).

    This software is provided under the Apache Software License.
    See the accompanying LICENSE file for more information.
"""

import argparse
import sys
import platform
import pathlib
from networksage_tools.converter import pcaputils
from networksage_tools.converter import zeekutils
from networksage_tools.common_utilities import utilities
from networksage_tools.common_utilities import dnsservice


def finish_conversion(dns, utils):
    """Handle final conversion of Zeek and PCAP inputs that don't depend on input-specific information.
    """
    # get passive DNS
    dns.get_passive_dns()
    # print("Found", len( utils.passive_dns_names_dict), "passive dns names")

    # use domain names instead of IP addresses wherever possible
    dns.map_destination_ips_to_names()

    # add current file's DNS to the PDNS file (for PCAPs).
    dns.update_passive_dns_repository()

    # save secFlows to file for hashing and to prepare final format
    utils.save_secflows_to_file()
    uuid = utils.get_random_uuid()
    utils.set_hash_value_for_sample(uuid)

    # capture secFlows in JSON format, and store with all information for final transmission
    utils.prepare_final_output_file()

    # check if output file looks sane
    utils.check_output_sanity()

    print("Cleaning up temporary files")
    utils.cleanup_files()

    if utils.output_dir is not None:
        # relocate the file
        utils.output_dir = pathlib.Path(utils.output_dir) # make sure it's a real Path object
        # create the directory if it doesn't exist
        utils.output_dir.mkdir(parents=True, exist_ok=True)
        start_path = pathlib.Path(utils.secflow_output_filepath)
        filename = start_path.stem + start_path.suffix
        utils.secflow_output_filepath = str(start_path.rename(pathlib.PurePath(utils.output_dir, filename)))
    print("Conversion complete! Final Secflow Output stored at", utils.secflow_output_filepath)


def convert_zeek(zeekfile_location, zeek_dnsfile_location=None, output_dir=None):
    """Handles all of the Zeek-specific conversion needs. Takes an optional DNS log.
    """
    my_platform = platform.system().lower()
    utils = utilities.Utilities(str(zeekfile_location), my_platform, output_dir=output_dir)  # create an instance of utils to use
    dns = dnsservice.DnsService(utils)  # create an instance of the DnsService class to use

    if zeek_dnsfile_location is not None:
        # store it if it exists
        dns.dns_logfile_path = str(zeek_dnsfile_location)

    # there's no filtered file in Zeek, so just grab name from original file.
    utils.filtered_filepath = utils.original_filepath

    # make sure the file is a valid Zeek conn log
    zeekutils.validate_file_format(utils)

    # most of the heavy lifting happens here
    success = zeekutils.zeek_2_secflows(utils)
    if success and len(utils.secflows) > 0:
        finish_conversion(dns, utils)
    else:
        print("No traffic was converted to Secflows.")
        utils.cleanup_files()


def convert_pcap(pcapfile_location, output_dir=None):
    """Handles all of the PCAP-specific conversion needs. Supports PCAPNG as well.
    """
    my_platform = platform.system().lower()
    utils = utilities.Utilities(str(pcapfile_location), my_platform, output_dir=output_dir)  # create an instance of utils to use
    dns = dnsservice.DnsService(utils)  # create an instance of the DnsService class to use

    # make sure the file is a valid capture file
    pcaputils.validate_file_format(utils)

    # most of the heavy lifting happens here
    pcaputils.pcap_to_secflow_converter(utils, dns)

    """ we only know how long a secFlow lasted when we've captured it all, so go back through the dictionary and update this now.
    """
    utils.set_secflow_durations()
    if len(utils.secflows) > 0:
        finish_conversion(dns, utils)
    else:
        print("No traffic was converted to Secflows.")
        utils.cleanup_files()



if __name__ == "__main__":
    # handle all arguments
    parser = argparse.ArgumentParser()
    zeek_group = parser.add_argument_group("zeek", "arguments available when analyzing Zeek files")
    zeek_group.add_argument("-z", "--zeekConnLog", help="a valid Zeek conn.log file", type=str)
    zeek_group.add_argument("-d", "--zeekDNSLog", help="a valid Zeek dns.log file")

    cap_group = parser.add_argument_group("pcap", "arguments available when analyzing capture (CAP, PCAP, PCAPNG) files")
    cap_group.add_argument("-p", "--pcap"
                          , help="indicates that a capture file (usually PCAP or PCAPNG) will be the input file to parse"
                          , type=str)

    args = parser.parse_args()

    pcapfile_location = None
    zeekfile_location = None
    zeek_dnsfile_location = None

    if args.zeekConnLog and args.pcap:
        """
        Check if both zeek and pcap files were inputted 
        """
        print("Error: can only parse Zeek OR PCAP, not both at same time.")
        sys.exit(1)
    elif args.zeekConnLog:
        zeekfile_location = pathlib.Path(args.zeekConnLog)
        if not zeekfile_location.is_file():
            print("Error:", zeekfile_location, "does not exist.")
            sys.exit(1)
        if args.zeekDNSLog:
            zeek_dnsfile_location = pathlib.Path(args.zeekDNSLog)
            if not zeek_dnsfile_location.is_file():
                print("Error:", zeek_dnsfile_location, "does not exist.")
                sys.exit(1)
        else:
            print("No Zeek DnsService log specified. Naming of secFlows may be suboptimal.")
            zeek_dnsfile_location = None
    elif args.pcap:
        pcapfile_location = pathlib.Path(args.pcap)
        if not pcapfile_location.is_file():
            print("Error:", pcapfile_location, "does not exist.")
            sys.exit(1)
    else:
        print("Missing -p or -z argument.")
        sys.exit(1)

    """
        Input type-specific file processing in this section
    """
    if args.zeekConnLog:
        convert_zeek(zeekfile_location, zeek_dnsfile_location)

    elif args.pcap:
        convert_pcap(pcapfile_location)


"""
    Copyright (c) 2021 David Pearson (david@seclarity.io)
    Date: 06/03/2021
    This file contains the Secflow class, which is used to capture the appropriate fields needed to create a Secflow object.

    This software is provided under the Apache Software License.
    See the accompanying LICENSE file for more information.
"""

import sys

class Secflow():
    def __init__(self, source_ip_port_string, s_port, d_ip, d_port, timing_info, file_start_time, protocol_information):
        self.key = source_ip_port_string
        self.source_port = s_port
        self.dest_port = d_port
        self.dest_ip = d_ip
        self.dest_name = ""  # where we'll store what we've learned the name to be
        self.min_ts = 0  # real value calculated in function at end
        self.max_ts = 0
        self.source_pkts = 0
        self.dest_pkts = 0
        self.source_payload_bytes = 0
        self.dest_payload_bytes = 0
        self.absolute_start_time = 0  # useful for figuring out which DNS name to use later (real value calculated in function at end)
        self.relative_start_time = 0  # how long after the file started did this session begin, rounded to 6 decimals
        self.duration = -1  # we can't know this until the end of session
        self.protocol_information = protocol_information # essentially should be empty unless it's ICMP right now
        self.calculate_timing_information(timing_info
                                        , file_start_time)  # we do this just once at the beginning, since these values are based on the start times

    def calculate_timing_information(self, timing_info, file_start_time):
        """Handles the fact that Packet-based Secflows will have a packet_header object with timing information, while
            Zeek-based flows will not. Populates min_ts, absolute_start_time, and relative_start_time fields.
        """

        if str(type(timing_info)) == "<class 'Pkthdr'>":
            ts_val = float(str(timing_info.getts()[0]) + "." + str(timing_info.getts()[1]).zfill(6))
            self.min_ts = ts_val
            self.absolute_start_time = ts_val
            self.relative_start_time = float("%.6f" % (ts_val - file_start_time))
        elif isinstance(timing_info, float):  # Zeek data should hit this case.
            self.min_ts = timing_info
            self.absolute_start_time = timing_info
            self.relative_start_time = float("%.6f" % (timing_info - file_start_time))
        else:
            print("Unrecognized instance passed in. Type:", type(timing_info))
            sys.exit(1)

    def set_secflow_duration(self):
        """Uses stored information to figure out how long this secflow lasted.
        """
        self.duration = float("%.6f" % (self.max_ts - self.min_ts))



"""
    Copyright (c) 2021 David Pearson (david@seclarity.io)
    Date: 06/03/2021
    This file contains utility functions that work with IP addresses.

    This software is provided under the Apache Software License.
    See the accompanying LICENSE file for more information.

    This file also leverages pcapy-ng, an Apache-licensed package that interfaces with the libpcap packet capture
    library. You can find more details about this repository at https://github.com/stamparm/pcapy-ng.
"""

import ipaddress
import pcapy

def check_if_local_ip(ip_address):
    """Helper function to figure out if the provided IP address is local, private, multicast, etc...
    """
    ip = ipaddress.ip_address(ip_address)
    return ip.is_private or ip.is_multicast or ip.is_loopback or ip.is_link_local  # we only want to work with globally-routable IPs, essentially

def collect_active_external_ips_from_capture_file(utils):
    """We want to know which IP addresses are actually active in a given input file, so we figure that out here. Active
       means that there was actually traffic to/from them, rather than just a DNS lookup that returned that IP
       address.
    """

    if len(utils.active_external_ips) > 0:
        print("Warning: there are already active external IPs stored. Results may differ from expectations.")
    with pcapy.open_offline(utils.filtered_filepath) as capfile:
        (hdr, pkt) = capfile.next()
        while hdr is not None:
            pkt_len = len(pkt)
            if pkt_len >= 34:  # length of IPv4 packet with standard header
                try:
                    if (pkt[0] == 0 and pkt[4:6] == b"\x00\x06" and pkt[14:16] == b"\x08\x00"):
                        #this IPv4 packet has a Linux-cooked header, so ignore first 2 bytes
                        pkt = pkt[2:]
                    ip_a = str(ipaddress.ip_address(pkt[26:30]))
                    ip_b = str(ipaddress.ip_address(pkt[30:34]))
                    if check_if_local_ip(ip_a):
                        pass
                    else:
                        utils.active_external_ips.add(ip_a)
                    if check_if_local_ip(ip_b):
                        pass
                    else:
                        utils.active_external_ips.add(ip_b)
                except:
                    print("IP addresses not found in range", pkt[26:34])
            hdr, pkt = capfile.next()



"""
    Copyright (c) 2021 David Pearson (david@seclarity.io)
    Date: 06/03/2021
    This file contains the DnsService class, which keeps the state of all DNS data for the creation of
    well-named Secflows. This class has helper functions to control passive naming from the incoming
    file we're analyzing and from recently seen sessions.

    This software is provided under the Apache Software License.
    See the accompanying LICENSE file for more information.
"""

import re
import socket
import pickle
import ipaddress
from pathlib import Path
from networksage_tools.common_utilities import iputils
from networksage_tools.common_utilities import packet
from networksage_tools.common_utilities import utilities

class DnsService:
    def __init__(self, utils):
        self.questions = dict()
        self.passive_dns_names_dict = dict()
        self.short_term_pdns_dict = dict() # saves passive DNS that was learned recently, but not necessarily during this streaming session
        self.active_dns_names = dict()
        self.unique_passive_dns_names = dict()  # results that only have one name for an IP in this file
        self.short_term_passive_dns_lock_name = "./shortTermPassiveDNS.pkl.lock"
        self.long_term_passive_dns_file_name = ""
        self.short_term_passive_dns_filename = ""
        self.dns_logfile_path = None
        if utils is None:  # shouldn't happen
            self.utils = utilities.Utilities("", "")
        else:
            self.utils = utils

    def order_passive_dns_by_timestamp(self):
        """Makes sure entries are sorted in order of timestamps
        """
        for entry in self.passive_dns_names_dict.keys():
            self.passive_dns_names_dict[entry] = sorted(self.passive_dns_names_dict[entry], key=lambda x: x[0])

    def get_unique_passive_dns_names(self):
        """Entries in this set only have one name (in this file) per IP address. This is used to update our
           long-term passive DNS file.
        """
        for entry in self.passive_dns_names_dict.keys():
            tmp_set = set()
            for tuple in self.passive_dns_names_dict[entry]:
                name = tuple[1]
                tmp_set.add(name)
            if len(tmp_set) == 1:
                self.unique_passive_dns_names[entry] = tmp_set.pop()  # grab the only name

    def parse_dns_records_from_dns_log(self):
        """Takes a Zeek dns.log, then creates and returns a dictionary of lists of tuples of packet source ports (for
           later filtering of local lookups), start times (for use later in figuring out which DNS name to use)
           and A or PTR record DNS resolutions found in the file. The resolutions are stored by name as follows:
              self.questions["someDomain.tld."]=[(someSrcPort, someTimestamp, "7.8.9.10"),(someSrcPort, someTimestamp, "1.2.3.4")]
              self.questions["4.3.2.1.in-addr.arpa."]=[(someSrcPort, someTimestamp, "someOtherDomain.tld.")]
           TODO: Handle PTR records for Zeek!
        """
        with open(self.dns_logfile_path, "r") as dns_logfile:
            for line in dns_logfile:
                if line.startswith("#"):  # ignore comment lines
                    continue
                dns_record = line.strip().split("\t")
                start_time = float(dns_record[0])
                roundtrip_time = dns_record[8]
                source_port = dns_record[3]
                q_name = dns_record[9] + "."  # we expect FQDNs in our files at this point
                q_class = dns_record[10]
                q_type = dns_record[12]
                rcode_name = dns_record[15]
                answers = dns_record[21]
                internet_a_record = False
                other_unknown_collectible_record = False
                if q_class == "1" and q_type == "1": # we only want successful DNS lookups to the Internet for A records.
                    internet_a_record = True
                elif roundtrip_time == "-" and rcode_name == "NOERROR": # these seem to occur when converting from PCAP to Zeek
                    other_unknown_collectible_record = True
                else: # everything else should be thrown away
                    continue

                if q_name not in self.questions.keys():
                    self.questions[q_name] = set()
                if answers.find(",") != -1:  # multiple answers
                    answers_list = answers.split(",")
                    for answer in answers_list:
                        try:
                            self.questions[q_name].add((source_port, start_time, str(ipaddress.ip_address(answer))))
                        except:  # it wasn't an IP address, so we'll skip collecting it
                            #print("Had an error with", answer)
                            continue
                else:
                    self.questions[q_name].add((source_port, start_time, answers))


    def parse_dns_records_from_capture_file(self):
        """Takes a capture file, then creates and returns a dictionary of lists of tuples of packet source ports (for
           use in filtering later), start times (for use later in figuring out which DNS name to use) and A or PTR
           record DNS resolutions found in the file. The resolutions are stored by name as follows:
              self.questions["someDomain.tld."]=[(someSrcPort, someTimestamp, "7.8.9.10"),(someSrcPort, someTimestamp, "1.2.3.4")]
              self.questions["4.3.2.1.in-addr.arpa."]=[(someSrcPort, someTimestamp, "someOtherDomain.tld.")]
           The parsing that occurs in this function is correlated with the DNS header, and (I believe) should
           handle both UDP and TCP DNS queries.
        """
        try:
            import pcapy
        except ImportError:
            print("Couldn't import pcapy")

        with pcapy.open_offline(self.utils.original_filepath) as capfile:
            capfile.setfilter("port domain")
            (hdr, pkt) = capfile.next()
            while hdr is not None:
                # we should only be getting DNS records now, so parse accordingly
                packet_info = packet.PacketInfo(pkt, hdr, self)  # prep everything for later use
                try:
                    is_resp = (pkt[packet_info.upper_layer_start + 2] > 127)  # highest bit would be 1, so must be above 127
                except:  # packet has no upper layer protocol (such as TCP SYN w/ no data)
                    hdr, pkt = capfile.next()
                    continue
                is_normal_query = (pkt[packet_info.upper_layer_start + 2] ^ 128 < 8)
                # bits 1-4 (0-based) of byte correspond to normal query (value should be 0)
                if is_resp and is_normal_query:  # only parse when needed
                    self.parse_local_lookup_from_packet(packet_info, pkt)
                else:  # skip query questions and/or non-normal responses
                    hdr, pkt = capfile.next()
                    continue
                hdr, pkt = capfile.next()


    def parse_local_lookup_from_packet(self, packet_info, pkt):
        txid = str(int.from_bytes(pkt[packet_info.upper_layer_start : packet_info.upper_layer_start + 2], "big"))
        num_qs = int.from_bytes(pkt[packet_info.upper_layer_start + 4 : packet_info.upper_layer_start + 6], "big")
        cnt = 0
        null_position = 0
        has_an_a_record = False
        has_ptr_record = False
        q_name = ""

        # there can be more than 1 question, so handle them all
        while cnt < num_qs:
            null_position = pkt[((packet_info.upper_layer_start + 13)
                                + (4 * cnt)
                                + null_position
                                ):].find(b"\x00")
            if int.from_bytes(pkt[(packet_info.upper_layer_start + 13) + null_position + 1: (packet_info.upper_layer_start + 13)
                                                                                            + null_position + 3]
                                , "big") == 1: # only do this for A records
                has_an_a_record = True
                if cnt == 0:
                    q_name=re.sub(r"[\x00-\x1f]"
                                , r"."
                                , pkt[(packet_info.upper_layer_start + 13): ((packet_info.upper_layer_start + 13)
                                                                            + null_position )].decode("ascii")) + "."
                    if q_name not in self.questions.keys():
                        self.questions[q_name] = set() # we don't yet have the answer
                else:
                    print("Multi-question DNS responses not currently implemented")
                    ###TODO. May not be terribly important in common usage, but should be handled at some point.
            elif int.from_bytes(pkt[(packet_info.upper_layer_start + 13)
                                    + null_position + 1 : (packet_info.upper_layer_start + 13) + null_position + 3]
                                , "big") == 12: # only do this for PTR records
                has_ptr_record = True
                if cnt == 0:
                    q_name=re.sub(r"[\x00-\x1f]",r".", pkt[(packet_info.upper_layer_start + 13):
                                                            ((packet_info.upper_layer_start + 13)
                                                            + null_position
                                                            )].decode("ascii")) + "."
                    if q_name not in self.questions.keys():
                        self.questions[q_name]= set() # we don't yet have the answer
                else:
                    print("Multi-question DNS responses not currently implemented")
                    ###TODO. May not be terribly important in common usage, but should be handled at some point.
            cnt+=1
        answers_start=((packet_info.upper_layer_start + 13)
                        + (4 * cnt)
                        + null_position
                        +1
                    ) # should be start of Answers section

        # find IP(s) for the A records and capture them
        if has_an_a_record:
            num_bytes_into_answers = 2
            while len(pkt) > (answers_start + num_bytes_into_answers):
                if num_bytes_into_answers == 2: # first time through
                    length_to_skip = 0
                else:
                    if length_to_skip == 0:
                        num_bytes_into_answers += (length_to_skip + 12)
                if pkt[answers_start + num_bytes_into_answers : answers_start + num_bytes_into_answers + 4] == b"\x00\x01\x00\x01": # an A record for an INternet address
                    # ignore TTL
                    try:
                        if int.from_bytes(pkt[answers_start + num_bytes_into_answers + 8 : answers_start + num_bytes_into_answers + 10]
                                        , "big") == 4: # it's an IPv4 A record
                            # add the information about that mapping as a tuple
                            self.questions[q_name].add(
                                                            (packet_info.source_port
                                                            , packet_info.packet_start_time
                                                            , str(ipaddress.ip_address(
                                                                    pkt[answers_start + num_bytes_into_answers + 10 : answers_start + num_bytes_into_answers + 14]
                                                                    )
                                                                )
                                                            )
                                                        )
                    except:
                        print("Some non-fatal error with", str(pkt[answers_start
                                                        + num_bytes_into_answers + 10
                                                        : answers_start
                                                        + num_bytes_into_answers + 14]
                                                    ),
                                                 "in", pkt)
                # calc how far until the next thing we should parse
                length_to_skip = int.from_bytes(pkt[answers_start + num_bytes_into_answers + 8 : answers_start + num_bytes_into_answers + 10]
                                            , "big")
                num_bytes_into_answers += (length_to_skip + 12)
        elif has_ptr_record:
            num_bytes_into_answers = 2
            while len(pkt) > (answers_start + num_bytes_into_answers):
                if num_bytes_into_answers == 2: # first time through
                    length_to_skip = 0
                    name_length = 0
                else:
                    if length_to_skip == 0:
                        num_bytes_into_answers += (length_to_skip + 12)
                if pkt[answers_start + num_bytes_into_answers : answers_start + num_bytes_into_answers + 4] == b"\x00\x0c\x00\x01": #a PTR record for an INternet address
                    #skip TTL
                    name_length = int.from_bytes(pkt[answers_start + num_bytes_into_answers + 8 : answers_start + num_bytes_into_answers + 10]
                                                , "big")
                    try:
                        answer=re.sub(r"[\x00-\x1f]"
                            , r"."
                            , pkt[answers_start + num_bytes_into_answers + 10 : answers_start + num_bytes_into_answers + 10 + name_length].decode("ascii")
                                        )[1:]
                    except UnicodeDecodeError as e:
                        #print("had an error decoding",pkt[answers_start+num_bytes_into_answers+10:answers_start+num_bytes_into_answers+10+name_length])
                        pass
                    self.questions[q_name].add((packet_info.source_port, packet_info.packet_start_time, answer))
                length_to_skip = name_length
                num_bytes_into_answers += (length_to_skip + 12)


    def get_passive_dns(self):
        """Takes the file we're analyzing and uses any DNS information contained in it to passively name as many
           of the external IP addresses as possible.
        """
        if not self.utils.zeekflows and self.utils.original_filepath:  # we're dealing with PCAP data
            # collect all of the active external IP addresses that need to be identified (if possible) passively.
            iputils.collect_active_external_ips_from_capture_file(self.utils)

            # collect all of the DNS records we see in our capture file
            self.parse_dns_records_from_capture_file()
        elif self.utils.zeekflows:  # this is Zeek data
            # collect all active external IP addresses that need to be identified
            for secflow in self.utils.secflows.keys():
                if not iputils.check_if_local_ip(str(self.utils.secflows[secflow].dest_ip)):
                    self.utils.active_external_ips.add(self.utils.secflows[secflow].dest_ip)

            # collect all DNS questions from DNS log, if it exists
            if self.dns_logfile_path is not None:
                self.parse_dns_records_from_dns_log()
        elif self.utils.is_streaming: #we're dealing with streaming data
            for secflow in self.utils.secflows.keys():
                if not iputils.check_if_local_ip(str(self.utils.secflows[secflow].dest_ip)):
                    self.utils.active_external_ips.add(self.utils.secflows[secflow].dest_ip)
        """Finally, for those IPs that are active, find the right DNS name and store it in a
           passive_dns_names_dict ordered dictionary keyed by stringified IP address with lists of tuples of resolution
           time and names as the values.
        """
        for name in self.questions.keys():
            for entry in self.questions[name]:
                dns_resolution = entry[2]
                resolution_time = entry[1]
                if dns_resolution in self.utils.active_external_ips:
                    if dns_resolution not in self.passive_dns_names_dict.keys():
                        self.passive_dns_names_dict[dns_resolution] = [(resolution_time, name[:-1])]
                    else:
                        self.passive_dns_names_dict[dns_resolution] += [(resolution_time, name[:-1])]


    def update_passive_dns_repository(self):
        """For all of the passive DNS from the current file, update the global passive DNS file if there
           are no current entries with the same IP address. If there are, delete that line from the passive DNS
           dictionary and do not save a new one.
        """

        long_term_pdns_path = Path(self.long_term_passive_dns_file_name)

        # iterate through current "ip name" for all DNS names learned passively from current file
        if not long_term_pdns_path.is_file():
            print("Warning:", long_term_pdns_path,
                  "does not exist.")
            with open(self.long_term_passive_dns_file_name, "w") as pdns_file:
                print("Created file now.") # just create the file
        with open(self.long_term_passive_dns_file_name, "r") as infile:
            # file is lines of "IP DNSname", so store in a dict with IP as key
            long_term_pdns = dict(line.strip().split(" ") for line in infile)
        self.get_unique_passive_dns_names()

        # search the global passive DNS file for IP-to-name conflicts
        for ip in self.unique_passive_dns_names.keys():
            if ip in long_term_pdns.keys():
                if self.unique_passive_dns_names[ip] == long_term_pdns[ip]:
                    continue  # still the same name, so do nothing
                else: # replace instead of delete, because on average the newer name will be more relevant temporally (since over time, most data will be fresh)
                    #print("Found conflict with name for IP", ip + ". Replacing.")
                    long_term_pdns[ip] = self.unique_passive_dns_names[ip]
            else:  # collect a newly-learned IP-to-name mapping
                long_term_pdns[ip] = self.unique_passive_dns_names[ip]

        # store the updated passive DNS file
        with open(self.long_term_passive_dns_file_name, "w") as outfile:
            for key, value in long_term_pdns.items():
                outfile.write("%s %s\n" % (key, value))

    def map_destination_ips_to_names(self):
        """Rewrite destination IPs (i.e. the thing on the Internet, generally) with names whenever possible.
           The strategy for doing this as correctly as possible is as follows:
                1. Replace any destination IPs that used passive DNS for the EXACT session (by time) from current
                   file.
                2. (streaming-specific) If a short-term passive DNS file exists, see if there is information in there
                   for the EXACT session (by time) we're currently trying to label. If so, label it.
                3. Replace any destination IPs with the most recent name we have from the current file (so if some IP
                   has multiple names for it in this file, we use the one that was most recently resolved at the time
                   we see this session.
                4. (streaming-specific) If a short-term passive DNS file exists, replace any destination IPs with the
                   most recent name we have. (NOTE: entries expire after ~5 minutes)
                5. Replace any destination IPs that are still unnamed using our long-term passive DNS collection,
                   which only stores names for IPs that we've never seen have more than one name.
                6. Replace any remaining destination IPs with what we learn from an active reverse DNS lookup.
                7. All remaining destination IPs stay as IP addresses that we attempt to lookup server-side.
        """

        self.long_term_passive_dns_file_name = "./longTermPassiveDNS.txt"
        try:
            long_term_pdns_file = Path(self.long_term_passive_dns_file_name)
        except:
            print("Warning:", long_term_pdns_file,
                              "does not exist. Will not use long-term passive DNS for this file.")
        long_term_pdns_dict = dict()
        self.short_term_passive_dns_filename = "./shortTermPassiveDNS.pkl"
        try:
            short_term_pdns_file = Path(self.short_term_passive_dns_filename)
            # the above we don't actually use when we're not streaming, but it's easier to instantiate the file and
            # ignore it than to rewrite the logic a little farther down.
            if self.utils.is_streaming: # only do this for streaming logic
                with open(short_term_pdns_file, "rb") as stpdf:
                    self.short_term_pdns_dict = pickle.load(stpdf)
                # merge short_term_pdns_file into the passive dns names dict so it can be sorted below:
                # get an iterator for this dict so we can pull its changes into the current passive dns names dict
                for entry in self.short_term_pdns_dict.keys():
                    if entry in self.passive_dns_names_dict.keys():
                        for val in self.short_term_pdns_dict[entry]:
                            if val not in self.passive_dns_names_dict[entry]:
                                self.passive_dns_names_dict[entry] += [val]
                    else:
                        self.passive_dns_names_dict[entry] = self.short_term_pdns_dict[entry]
        except:
            if self.utils.is_streaming: # only do this for streaming logic
                print("Short-term passive DNS file", short_term_pdns_file
                    , "does not exist (or some other error). Will not use short-term knowledge for this iteration.")

        self.order_passive_dns_by_timestamp()  # make sure we're sorted by timestamp

        if not long_term_pdns_file.is_file():
            print("Warning:", long_term_pdns_file,
                  "does not exist. Will not use long-term passive DNS for this file.")
        else:
            with open(self.long_term_passive_dns_file_name) as infile:
                # file is lines of "IP DNSname", so store in a dict with IP as key
                long_term_pdns_dict = dict(line.strip().split(" ") for line in infile)

        entry = None # entry is only used for streaming logic
        with open(short_term_pdns_file, "wb") as stpdf_handle: # we actually do plan to overwrite it every time so that we can keep updates sanely
            for secflow in self.utils.secflows.keys():
                named = False  # keep track of whether we've named something each time
                flow_ip = self.utils.secflows[secflow].dest_ip  # local var to make things less verbose in this loop
                if flow_ip in self.passive_dns_names_dict.keys():
                    try:
                        for tuple in list(reversed(self.passive_dns_names_dict[flow_ip])):
                            """The list of tuples is already reverse ordered by timestamp so the first time that is less
                            is the most correct one to use here. """
                            resolution_timestamp = tuple[0]  # grab timestamp
                            name = tuple[1]
                            if resolution_timestamp <= self.utils.secflows[secflow].absolute_start_time:
                                self.utils.secflows[secflow].dest_name = name
                                self.utils.secflows[secflow].destination_name_source = "passive"
                                named = True
                                entry = (resolution_timestamp, name)
                                break  # quit as soon as we get one
                            else:
                                continue
                    except:
                        print("Error: IP address", flow_ip, "has run out of passive_dns_names_dict.")
                        continue
                if not named and self.utils.secflows[secflow].dest_ip in long_term_pdns_dict.keys():
                    # replace any possible destinations with passive DNS from overall PDNS file, if it exists.
                    self.utils.secflows[secflow].dest_name = long_term_pdns_dict[flow_ip]
                    self.utils.secflows[secflow].destination_name_source = "cache"
                    named = True
                if not named:
                    if iputils.check_if_local_ip(flow_ip):
                        print("THIS SHOULDN'T HAPPEN! Info:")
                        self.utils.secflows[secflow].dest_name = flow_ip
                        self.utils.secflows[secflow].destination_name_source = "original"
                        print("\tSource:", self.utils.secflows[secflow].key)
                        print("\tDestination:", self.utils.secflows[secflow].dest_ip+":"+self.utils.secflows[secflow].dest_port)
                        continue  # this should not happen...
                    try:  # if there is something in our current file, grab first one.
                        # print("Deciding to label", flow_ip, "with", self.passive_dns_names_dict[flow_ip][0][1])
                        self.utils.secflows[secflow].dest_name = self.passive_dns_names_dict[flow_ip][0][1]
                        self.utils.secflows[secflow].destination_name_source = "passive"
                        entry = (self.passive_dns_names_dict[flow_ip][0][0], self.passive_dns_names_dict[flow_ip][0][1])
                        named = True
                    except:  # if not, just keep the IP
                        self.utils.secflows[secflow].dest_name = flow_ip
                        self.utils.secflows[secflow].destination_name_source = "original"
                if self.utils.is_streaming: # only do this for streaming logic
                    if named and entry is not None: # we had picked up our name from current file's passive DNS
                        # capture the particular entry in the short-term passive DNS file
                        if flow_ip not in self.short_term_pdns_dict.keys():
                            """If a name doesn't exist in the current version of the short-term dns names dict, we should
                               save it so we can write it to the file.
                            """
                            self.short_term_pdns_dict[flow_ip] = [(entry[0], entry[1])]
                        else:
                            """If a name DOES exist in the current version of the short-term dns names dict (but it doesn't
                               contain the exact timestamp we currently have), we should UPDATE it so we can write it to
                               the file.
                            """
                            if (entry[0], entry[1]) not in self.short_term_pdns_dict[flow_ip]:
                                self.short_term_pdns_dict[flow_ip] += [(entry[0], entry[1])]
                continue
            if self.utils.is_streaming: # only do this for streaming logic
                pickle.dump(self.short_term_pdns_dict, stpdf_handle) # at the end, (over)write pickle file with our current knowledge


    def collect_local_lookups(self):
        """This function parses the questions dictionary and collects the destination ports and timestamps for all of
           the lookups that are either local forward lookups:
               questions["hello.local."]=[("44332", "1234.5678", "10.17.18.24")]
           or local reverse lookups:
               questions["5.15.168.192.in-addr.arpa."]=[("55117", "2345.6789", "hi.local.")]
           These are collected in a list and returned to the caller.

           Some testing examples for local forward and local reverse lookups:
           questions["hello.local."]=[("51317", "1234.5678", "10.17.18.24"),("54217", "1235.7789", "172.16.18.40")]
           questions["5.15.168.192.in-addr.arpa."]=[("8675", "1245.5432", "hi.local.")]
        """

        local_lookups = []
        names = self.questions.keys()
        for name in names:
            if len(self.questions[name]) == 0:
                continue
            for entry in self.questions[name]:  # iterate through all resolutions for a given DNS name
                dest_port = entry[0]
                resolution = entry[2]
                # remove local forward lookups
                if re.match(self.utils.local_ips_forward_regex, resolution):
                    local_lookups += [dest_port + " or "]
                # remove local reverse lookups
                else:
                    if re.match(self.utils.local_ips_reverse_regex, name):
                        local_lookups += [dest_port + " or "]
        # remove extra " or " from end of list
        if len(local_lookups) > 0:
            last = local_lookups.pop()
            local_lookups += [last[:-4]]
        return local_lookups



"""
    Copyright (c) 2021 David Pearson (david@seclarity.io)
    Date: 06/02/2021
    This file contains utilities that allow a Zeek conn (and an optional dns) log to be converted into equivalent
    secFlows.

    This software is provided under the Apache Software License.
    See the accompanying LICENSE file for more information.
"""

import re
import sys
import magic
from networksage_tools.common_utilities import iputils
from networksage_tools.common_utilities import secflow
from networksage_tools.converter import zeekflow


def remove_local_traffic_from_zeek_file(utils):
    """Removes the local traffic from the incoming Zeek file dictionary.
    """
    tmp_dict = dict()
    for flow_group in utils.zeekflows.keys():
        tmp_flowlist = []  # clear it each time through the group
        for flow in utils.zeekflows[flow_group]:
            if iputils.check_if_local_ip(flow.source_ip) and iputils.check_if_local_ip(flow.dest_ip):
                # print("removing", flow.source_ip, "<->", flow.dest_ip, "from flows dictionary.")
                continue
            else:  # one of the sides is not local
                tmp_flowlist += [flow]
        tmp_dict[flow_group] = tmp_flowlist  # keep only the ones that aren't local-to-local
    utils.zeekflows = tmp_dict


def store_zeekflows(utils):
    """Takes a Zeek file, removes the header lines, and stores the input as a dictionary. We store each line in the file
       as a flow, but then we store each flow object by the key used for secFlows.
    """

    with open(utils.original_filepath) as infile:
        for line in infile:
            if line.startswith("#"):  # it's a comment line
                continue
            flowdata = line.strip().split("\t")
            zeekflow_object = zeekflow.ZeekFlow(flowdata)  # an individual line, which actually doesn't always capture all of the data for a 4-tuple.
            if zeekflow_object.secflow_key not in utils.zeekflows.keys():
                utils.zeekflows[zeekflow_object.secflow_key] = []
            utils.zeekflows[zeekflow_object.secflow_key] += [(zeekflow_object)]


def identify_zeekfile_start_time(utils):
    """Takes a Zeek flows dictionary and finds the earliest time.
    """
    ts = float(99999999999)
    for flow_group in utils.zeekflows.keys():
        for flow in utils.zeekflows[flow_group]:
            ts = min(ts, float(flow.timestamp))
    utils.file_start_time = ts

    # print("File start time:", self.file_start_time)


def convert_zeek_to_secflow(utils):
    """This function takes the original Zeek conn log flows and converts them into secFlows. secFlows may be made up of
       one or more Zeek conn log flows, since Zeek parses out service-related communication between two IP:Port pairs
       separately from the underlying transport-layer (i.e. TCP/UDP) flow. In the logic below, those multiple conn log
       flows are grouped together into a flow group. Generally speaking, this should guarantee that the destination IP
       and port are the same for all of the flows in this group, since source ports should be incrementally increasing
       on the host. It is certainly possible that an adversary could cause the system to use the same source port for
       more than one session, which would cause issues here and in our PCAP analysis. Handling those cases (perhaps by
       using a key requiring the unique 4-tuple of sourceIP:source_port+dest_ip:dest_port) is future work.
    """

    for flow_group in utils.zeekflows.keys():
        # initialize a few variables to keep track of state inside flow groups
        max_nonservice_sb = 0
        max_service_sb = 0
        max_nonservice_db = 0
        max_service_db = 0

        # sort the flows in a flow group by their start timestamp
        utils.zeekflows[flow_group] = sorted(utils.zeekflows[flow_group], key=lambda x: x.timestamp)

        secflow_object = None  # start off empty

        for zeekflow in utils.zeekflows[flow_group]:
            if iputils.check_if_local_ip(str(zeekflow.source_ip)) and iputils.check_if_local_ip(str(zeekflow.dest_ip)):
                break  # we don't want local-to-local traffic!
            if iputils.check_if_local_ip(str(zeekflow.dest_ip)) and not iputils.check_if_local_ip(str(zeekflow.source_ip)):
                print("We think Zeek is wrong in identifying who the source is. Flipping order.")
                zeekflow.flip_zeek_order() # flip things related to bytes and directionality
                secflow_object = secflow.Secflow(zeekflow.secflow_key
                                                , zeekflow.source_port
                                                , zeekflow.dest_ip
                                                , zeekflow.dest_port
                                                , float(zeekflow.timestamp)
                                                , float(utils.file_start_time)
                                                , zeekflow.protocol_information
                                                )  # create a new secFlow object
                utils.secflows[zeekflow.secflow_key] = secflow_object  # store it
            if zeekflow.secflow_key not in utils.secflows.keys():  # we've not collected this sourceIP:source_port
                secflow_object = secflow.Secflow(zeekflow.secflow_key
                                                , zeekflow.source_port
                                                , zeekflow.dest_ip
                                                , zeekflow.dest_port
                                                , float(zeekflow.timestamp)
                                                , float(utils.file_start_time)
                                                , zeekflow.protocol_information
                                                )  # create a new secFlow object
                utils.secflows[zeekflow.secflow_key] = secflow_object  # store it

            # collect packet info from Zeek flow
            secflow_object.source_pkts += int(zeekflow.source_pkts)
            secflow_object.dest_pkts += int(zeekflow.dest_pkts)

            # get the max duration of the flow
            if zeekflow.duration != "-":
                secflow_object.max_ts = max(secflow_object.max_ts, zeekflow.timestamp + float(zeekflow.duration))
            else:
                secflow_object.max_ts = max(secflow_object.max_ts, zeekflow.timestamp)
            # get correct byte counts (according to our secFlow calcuations)
            if zeekflow.service == "-":  # no service identified for this particular flow
                if zeekflow.history == "S":  # and we have ONLY a SYN flag
                    max_nonservice_sb = 0
                    max_nonservice_db = 0
                elif max_nonservice_sb == 0 or max_nonservice_db == 0:
                    if not re.search(r"d", zeekflow.history):  # Zeek saw no payload bytes from the source
                        if not re.search(r"t", zeekflow.history):  # Zeek saw no retransmitted payloads from the source
                            max_nonservice_sb = 0
                        else:
                            if zeekflow.source_bytes == "-": # it's not a protocol that contains bytes in upper layers
                                max_nonservice_sb = 0
                            else:
                                max_nonservice_sb = int(zeekflow.source_bytes)
                    else:
                        if zeekflow.source_bytes == "-": # it's not a protocol that contains bytes in upper layers
                            max_nonservice_sb = 0
                        else:
                            max_nonservice_sb = int(zeekflow.source_bytes)
                    if not re.search(r"D", zeekflow.history):  # Zeek saw no payload bytes from the destination
                        if not re.search(r"t",
                                         zeekflow.history):  # Zeek saw no retransmitted payloads from the destination
                            max_nonservice_db = 0
                        else:
                            max_nonservice_db = int(zeekflow.dest_bytes)
                    else:
                        max_nonservice_db = int(zeekflow.dest_bytes)
                else:
                    max_nonservice_sb = max(max_nonservice_sb, int(zeekflow.source_ip_bytes))
                    max_nonservice_db = max(max_nonservice_db, int(zeekflow.dest_ip_bytes))
            else:
                if max_service_sb == 0 or max_service_db == 0:
                    max_service_sb = int(zeekflow.source_bytes)
                    max_service_db = int(zeekflow.dest_bytes)
                else:
                    max_service_sb = max(max_service_sb, int(zeekflow.source_bytes))
                    max_service_db = max(max_service_db, int(zeekflow.dest_bytes))
        if secflow_object is not None:
            # at end of the flow group, collect what we've captured and update the secFlow object if the value is higher than what we already have
            secflow_object.source_payload_bytes = max(secflow_object.source_payload_bytes
                                                    , max(max_nonservice_sb, max_service_sb))
            secflow_object.dest_payload_bytes = max(secflow_object.dest_payload_bytes
                                                    , max(max_nonservice_db, max_service_db))
            secflow_object.set_secflow_duration()


def validate_file_format(utils):
    """Validate if file is an ASCII text file -- very weak check for Zeek, unfortunately.
    """
    if not re.match(r"^ASCII text$", magic.from_file(utils.original_filepath)):
        print("Error:", utils.original_filepath + ",", "of type", magic.from_file(utils.original_filepath),
              "is not an accepted file type.")
        sys.exit(1)


def zeek_2_secflows(utils):
    """Given a conn log that has been lightly validated, store the Zeek flows, remove any local traffic and convert the
       Zeek flows to secFlows.
    """

    # store file as a dict for easier processing
    store_zeekflows(utils)
    if len(utils.zeekflows) == 0:
        print("No Zeek flows were found in file. No traffic was converted to Secflows.")
        return False

    # remove local connections
    remove_local_traffic_from_zeek_file(utils)

    # get file start time
    identify_zeekfile_start_time(utils)

    # convert from Zeek to secFlow
    convert_zeek_to_secflow(utils)
    return True



"""
    Copyright (c) 2021 David Pearson (david@seclarity.io)
    Date: 06/03/2021
    This file contains the PacketInfo class, which allows us to take a network packet as input and calculate all
    relevant information needed to make correct decisions on IP, TCP, and UDP data.

    This software is provided under the Apache Software License.
    See the accompanying LICENSE file for more information.
"""

import ipaddress
from networksage_tools.common_utilities import iputils
from networksage_tools.common_utilities import secflow

class PacketInfo():
    def __init__(self, pkt, packet_header, utils):
        """Store/Calculate relevant fields
        """
        self.packet_start_time = float(str(packet_header.getts()[0]) + "." + str(packet_header.getts()[1]).zfill(
            6))  # unix seconds.microseconds format
        self.is_ipv4_by_ethernet_layer = (pkt[12:14] == b"\x08\x00")
        self.is_ipv4_by_ip_layer = (((pkt[14] ^ 64) >> 4) == 0)
        self.ip_layer_header_length = ((pkt[14] ^ 64) * 4)
        self.ip_layer_length_field = int.from_bytes(pkt[16:18], "big")
        self.transport_protocol = pkt[23]  # 6 for TCP, 17 for UDP
        self.source_ip = str(ipaddress.ip_address(pkt[26:30]))
        self.dest_ip = str(ipaddress.ip_address(pkt[30:34]))
        self.transport_layer_start = 34 + self.ip_layer_header_length - 20
        self.source_port = str(int.from_bytes(pkt[self.transport_layer_start : self.transport_layer_start + 2], "big"))
        self.dest_port = str(int.from_bytes(pkt[self.transport_layer_start + 2 : self.transport_layer_start + 4], "big"))
        self.protocol_information = ""
        if self.transport_protocol == 6:  # TCP
            self.tcp_header_length = ((pkt[self.transport_layer_start + 12] >> 4) * 4)
            self.upper_layer_length = self.ip_layer_length_field - self.ip_layer_header_length - self.tcp_header_length
            self.upper_layer_start = 14 + self.ip_layer_header_length + self.tcp_header_length
        elif self.transport_protocol == 17:  # UDP
            self.udp_full_length = int.from_bytes(pkt[self.transport_layer_start + 4 : self.transport_layer_start + 6], "big")
            self.upper_layer_length = self.udp_full_length - 8  # 8 is always the length of the UDP header
            self.upper_layer_start = 14 + self.ip_layer_header_length + 8
        elif self.transport_protocol == 1: # ICMP, which doesn't have a transport layer
            self.source_port = ""
            self.dest_port = ""
            self.protocol_information = "ICMP"
            self.transport_layer_length = 0
            self.upper_layer_length = 0
        else:  # something is wrong
            self.transport_layer_length = -1
            self.upper_layer_length = -1
        if utils is None:  # shouldn't happen
            self.utils = utilities.Utilities("", " ")
        else:
            self.utils = utils

    def determine_session_directionality(self, hdr):
        """Determines which side of a given not-yet-identified session is the source vs. the destination based on the
            provided packet. Incoming data will look like the following:
                self.source_ip: "1.2.3.4"
                self.dest_ip: "8.7.6.5"
                self.source_port: "55"
                self.dest_port: "3333"
            Even though the information is labeled source and dest, we don't actually know which is the source/dest for
            the entire session (this data is labeled as source/dest from the CURRENT PACKET). Therefore, the source
            items are relabeled as the "A" side of the session to which this packet belongs, while the dest items are
            the "B" side. Those packets (if written out in a line) would look something like:
                ip_a:port_a > ip_b:port_b <some other data>

            The directionality logic comes from knowledge of relevant RFCs plus experience with common session behavior
            on the Internet.
        """

        side_a = self.source_ip + ":" + self.source_port
        side_b = self.dest_ip + ":" + self.dest_port

        if iputils.check_if_local_ip(str(self.source_ip)):
            if iputils.check_if_local_ip(str(self.dest_ip)):
                # both sides are local; should only occur when it's DNS that we needed earlier
                return None
            if self.source_port == "" and self.protocol_information in ["ICMP"]:
                """ICMP is treated specially, since it doesn't have a port number...so we need to save the whole
                   source/dest IP pair in addition to the protocol information. Additionally, we already know that the
                   source IP of this packet is local, so we'll set it as the source
                """
                object = secflow.Secflow(side_a + self.protocol_information + "-" + side_b + self.protocol_information
                                        , self.source_port
                                        , self.dest_ip
                                        , self.dest_port
                                        , hdr
                                        , self.utils.file_start_time
                                        , self.protocol_information
                                        )
                self.utils.secflows[side_a + self.protocol_information + "-" + side_b + self.protocol_information] = object
                return self.utils.secflows[side_a + self.protocol_information + "-" + side_b + self.protocol_information]
            if int(self.source_port) > 49151:
                """Source IP is local and has an ephemeral port, so create a new secflow object, store it in our
                    Secflows dictionary, and return it.
                """
                object = secflow.Secflow(side_a
                                        , self.source_port
                                        , self.dest_ip
                                        , self.dest_port
                                        , hdr
                                        , self.utils.file_start_time
                                        , self.protocol_information
                                        )
                self.utils.secflows[side_a] = object
                return self.utils.secflows[side_a]
            else:  # first IP has well-known or registered port, meaning it may actually be dest...so check the other side!
                if iputils.check_if_local_ip(str(self.dest_ip)):
                    if int(self.dest_port) > 49151:
                        """This IP is local and has an ephemeral port, so create a new secflow object, store it in our
                            Secflows dictionary, and return it.
                        """
                        object = secflow.Secflow(side_b
                                                , self.dest_port
                                                , self.source_ip
                                                , self.source_port
                                                , hdr
                                                , self.utils.file_start_time
                                                , self.protocol_information
                                                )
                        self.utils.secflows[side_b] = object
                        return self.utils.secflows[side_b]
                    else:  # both sides have ephemeral ports, and we see this packet first. So just call the first IP the source for now...shouldn't happen in this code!
                        print("Warning: found two local IPs. This shouldn't happen!")
                        object = secflow.Secflow(side_a
                                                , self.source_port
                                                , self.dest_ip
                                                , self.dest_port
                                                , hdr
                                                , self.utils.file_start_time
                                                , self.protocol_information
                                                )
                        self.utils.secflows[side_a] = object
                        return self.utils.secflows[side_a]
                else:  # possible source (2nd IP in line) is not local and may have a well-known or registered port
                    if int(self.dest_port) > 49151:  # possibleSource is NOT local and has ephemeral port
                        object = secflow.Secflow(side_b
                                                , self.dest_port
                                                , self.source_ip
                                                , self.source_port
                                                , hdr
                                                , self.utils.file_start_time
                                                , self.protocol_information
                                                )
                        self.utils.secflows[side_b] = object
                        return self.utils.secflows[side_b]
                    """Here we already know that this packet's source IP is local, the dest IP is not local, the source
                       port is not ephemeral, and the dest port is not ephemeral...so now we should determine if the
                       destination's port is well-known. If it is, then it's the destination. If it's not, then we
                       should default to making the local IP the source.
                    """
                    if int(self.dest_port) < 1024:
                        object = secflow.Secflow(side_a
                                                , self.source_port
                                                , self.dest_ip
                                                , self.dest_port
                                                , hdr
                                                , self.utils.file_start_time
                                                , self.protocol_information
                                                )
                        self.utils.secflows[side_a] = object
                        return self.utils.secflows[side_a]
                    else:
                        """Both have ephemeral ports, and we see this packet first, but the other side is a local IP.
                             So just call the local IP the source for now.
                        """
                        object = secflow.Secflow(side_b
                                                , self.dest_port
                                                , self.source_ip
                                                , self.source_port
                                                , hdr
                                                , self.utils.file_start_time
                                                , self.protocol_information
                                                )
                        self.utils.secflows[side_b] = object
                        return self.utils.secflows[side_b]
        else:  # first IP is not a local IP, so check the second IP in the line
            """ICMP is treated specially, since it doesn't have a port number...so we need to save the whole
               source/dest IP pair in addition to the protocol information.
            """
            if self.dest_port == "" and self.protocol_information in ["ICMP"]:
                if not iputils.check_if_local_ip(str(self.dest_ip)):
                    """Neither IP for this packet is local, and we saw this packet first, so we'll set it as the source.
                    """
                    object = secflow.Secflow(side_a + self.protocol_information + "-" + side_b + self.protocol_information
                                            , self.source_port
                                            , self.dest_ip
                                            , self.dest_port
                                            , hdr
                                            , self.utils.file_start_time
                                            , self.protocol_information
                                            )
                    self.utils.secflows[side_a + self.protocol_information + "-" + side_b + self.protocol_information] = object
                    return self.utils.secflows[side_a + self.protocol_information + "-" + side_b + self.protocol_information]
                else:
                    object = secflow.Secflow(side_b + self.protocol_information + "-" + side_a + self.protocol_information
                                            , self.dest_port
                                            , self.source_ip
                                            , self.source_port
                                            , hdr
                                            , self.utils.file_start_time
                                            , self.protocol_information
                                            )
                    self.utils.secflows[side_b] = object
                    return self.utils.secflows[side_b]
            if int(self.dest_port) > 49151:
                """ The second IP at least has an ephemeral port, so consider it as the source, create a new object,
                    store it in our Secflows dictionary, and return it.
                """
                object = secflow.Secflow(side_b
                                        , self.dest_port
                                        , self.source_ip
                                        , self.source_port
                                        , hdr
                                        , self.utils.file_start_time
                                        , self.protocol_information
                                        )
                self.utils.secflows[side_b] = object
                return self.utils.secflows[side_b]
            if not iputils.check_if_local_ip(str(self.dest_ip)):
                """The second IP is not local and has a port in the well-known range, so consider it as the
                   destination, create a new object, store it in our Secflows dictionary, and return it.
                """
                if int(self.dest_port) < 1024:
                    object = secflow.Secflow(side_a
                                            , self.source_port
                                            , self.dest_ip
                                            , self.dest_port
                                            , hdr
                                            , self.utils.file_start_time
                                            , self.protocol_information
                                            )
                    self.utils.secflows[side_a] = object
                    return self.utils.secflows[side_a]
                else:
                    print("We don't handle this case! Info:")
                    print("Info (from determine_session_directionality):", self.dest_ip + ":" + self.dest_port
                        , "listed as dest")
                    print(self.source_ip + ":" + self.source_port
                        , "listed as source")
            else:
                """At least the second IP has a non-ephemeral port, and we see this packet first, but the second IP is
                   a local IP. So just call the second IP the source for now.
                """
                object = secflow.Secflow(side_b
                                        , self.dest_port
                                        , self.source_ip
                                        , self.source_port
                                        , hdr
                                        , self.utils.file_start_time
                                        , self.protocol_information
                                        )
                self.utils.secflows[side_b] = object
                return self.utils.secflows[side_b]

    def is_local_dns_response(self, packet_data):
        try:
            is_resp = (packet_data[self.upper_layer_start + 2] > 127) # highest bit would be 1, so must be above 127
            is_normal_query = (packet_data[self.upper_layer_start + 2] ^ 128 < 8) # bits 1-4 (0-based) of byte correspond to normal query (value should be 0)
            if is_normal_query and is_resp:
                return True
            elif is_normal_query:
                return False # it is a DNS query, but it's not the response (which is what we need)
        except:
            return False



"""
    Copyright (c) 2022 David Pearson (david@seclarity.io)
    Date: 02/03/2022
    This file contains tests for the converter code.

    This software is provided under the Apache Software License.
    See the accompanying LICENSE file for more information.
"""

import pathlib
import importlib.resources
from networksage_tools.converter import convert

def run_tests():
    test_cases = [
                    { "inputType": "pcapng"
                    , "filename": "testCase1_chromeWin10StartChromeBrowser.pcapng"
                    }
                    ,
                    { "inputType": "pcapng"
                    , "filename": "testCase2_chromeWin10VisitAmazonScamaDotCom.pcapng"
                    }
                    ,
                    { "inputType": "pcapng"
                    , "filename": "testCase3_onePacketNotProcessable.pcapng"
                    }
                    ,
                    { "inputType": "pcapng"
                    , "filename": "testCase4_onePacketIsProcessable.pcapng"
                    }
                    ,
                    { "inputType": "pcapng"
                    , "filename": "testCase5_oneFlowPlusDNSIsProcessable.pcapng"
                    }
                    ,
                    { "inputType": "pcap"
                    , "filename": "testCase6_icmp_plus_others.pcap"
                    }
                    ,
                    { "inputType": "zeek"
                    , "filename": "testCase7_chromeWin10StartChromeBrowser.conn.log"
                    , "hasDNS": True
                    }
                    ,
                    { "inputType": "zeek"
                    , "filename": "testCase8_chromeWin10VisitAmazonScamaDotCom.conn.log"
                    , "hasDNS": True
                    }
                    ,
                    { "inputType": "zeek"
                    , "filename": "testCase9_oneFlowNotProcessable.conn.log"
                    , "hasDNS": False
                    }
                    ,
                    { "inputType": "zeek"
                    , "filename": "testCase10_oneFlowIsProcessable.conn.log"
                    , "hasDNS": False
                    }
                    ,
                    { "inputType": "zeek"
                    , "filename": "testCase11_emptySansHeaders.conn.log"
                    , "hasDNS": False
                    }
                    ,
                    { "inputType": "zeek"
                    , "filename": "testCase12_icmp_plus_others.conn.log"
                    , "hasDNS": True
                    }
                ]
    test_dir = None
    with importlib.resources.path(__package__, "tests") as p:
        test_dir = p
    inputs_location = str(pathlib.PurePath(test_dir, "inputs"))
    outputs_location = pathlib.PurePath(test_dir, "outputs")
    expected_outputs_location = pathlib.PurePath(test_dir, "expectedOutputs")

    for test in test_cases:
        try:
            if test["inputType"] in ["pcap", "pcapng"]:
                print("Converting", test["filename"], "from PCAP(NG) to Secflow.")
                convert.convert_pcap(inputs_location+"/"+test["filename"]
                                    , output_dir=outputs_location
                                    )
            elif test["inputType"] == "zeek":
                print("Converting", test["filename"], "from Zeek to Secflow.")
                if "hasDNS" in test.keys() and test["hasDNS"]:
                    dns_name = test["filename"].replace(".conn.log", ".dns.log")
                    convert.convert_zeek(inputs_location+"/"+test["filename"]
                                        , zeek_dnsfile_location=inputs_location+"/"+dns_name
                                        , output_dir=outputs_location
                                        )
                else:
                    convert.convert_zeek(inputs_location+"/"+test["filename"]
                                        , output_dir=outputs_location
                                        )
            else:
                print("Unrecognized input type", test["inputType"], "...skipping!")
                continue
            print("===============================================================")
        except:
            print("Something unexpected happened for this test case. Skipping.")
    print("Tests complete! Compare tests in", outputs_location, "to those in", expected_outputs_location, "to determine if any issues have occurred.")


"""
    Copyright (c) 2021 David Pearson (david@seclarity.io)
    Date: 06/03/2021
    This file contains the ZeekFlow class, which captures the fields that are found in a Zeek flow from the conn.log
    file. Formatting should always look like the following ("field #" lines are my annotation):

        Field #  1       2        3                 4               5
                ts      uid     id.orig_h       id.orig_p       id.resp_h
        Field #    6             7        8        9               10
                id.resp_p       proto   service duration        orig_bytes
        Field #     11              12              13              14
                resp_bytes      conn_state      local_orig      local_resp
        Field #     15             16      17               18             19
                missed_bytes    history orig_pkts       orig_ip_bytes  resp_pkts
        Field #       20              21
                resp_ip_bytes   tunnel_parents

    An example of a flow line from the conn log is as follows ("field #" lines are my annotation):

        Field #       1                        2                      3
                1601060272.439360       CC9S3G178KjzSMTGRk      192.168.100.224
        Field #   4             5        6       7       8          9
                 137    192.168.100.255 137     udp     dns     12.114023
        Field #  10     11    12      13     14      15      16       17
                1186    0     S0      -       -       0       D       23
        Field #  18     19      20      21
                1830    0       0       -

    This software is provided under the Apache Software License.
    See the accompanying LICENSE file for more information.
"""

import copy

class ZeekFlow():
    def __init__(self, flowdata):
        self.secflow_key = flowdata[2] + ":" + flowdata[3]  # useful for later conversion, but also use it here.
        self.timestamp = float(flowdata[0])
        self.unique_id = flowdata[1]
        self.source_ip = flowdata[2]
        self.source_port = flowdata[3]
        self.dest_ip = flowdata[4]
        self.dest_port = flowdata[5]
        self.trans_proto = flowdata[6]
        self.protocol_information = ""
        if self.trans_proto == "icmp":
            self.source_port = ""
            self.dest_port = ""
            self.protocol_information = "ICMP"
            self.secflow_key = self.source_ip+":"+self.protocol_information+"-"+self.dest_ip+":"+self.protocol_information
        self.service = flowdata[7]
        self.duration = flowdata[8]
        self.source_bytes = flowdata[9]
        self.dest_bytes = flowdata[10]
        self.connection_state = flowdata[11]
        self.local_orig = flowdata[12]
        self.local_resp = flowdata[13]
        self.missed_bytes = flowdata[14]
        self.history = flowdata[15]
        self.source_pkts = flowdata[16]
        self.source_ip_bytes = flowdata[17]
        self.dest_pkts = flowdata[18]
        self.dest_ip_bytes = flowdata[19]
        self.tunnel_parents = flowdata[20]

    def flip_zeek_order(self):
        orig_zeek_flow = copy.deepcopy(self)
        if self.protocol_information != "ICMP":
            self.secflow_key = orig_zeek_flow.dest_ip + ":" + orig_zeek_flow.dest_port
        else:
            self.secflow_key = orig_zeek_flow.dest_ip+":"+self.protocol_information+"-"+orig_zeek_flow.source_ip+":"+self.protocol_information
        self.source_ip = orig_zeek_flow.dest_ip
        self.source_port = orig_zeek_flow.dest_port
        self.dest_ip = orig_zeek_flow.source_ip
        self.dest_port = orig_zeek_flow.source_port
        self.source_bytes = orig_zeek_flow.dest_bytes
        self.dest_bytes = orig_zeek_flow.source_bytes
        self.history = orig_zeek_flow.history.swapcase()
        self.source_pkts = orig_zeek_flow.dest_pkts
        self.source_ip_bytes = orig_zeek_flow.dest_ip_bytes
        self.dest_pkts = orig_zeek_flow.source_pkts
        self.dest_ip_bytes = orig_zeek_flow.source_ip_bytes


"""
    Copyright (c) 2021 David Pearson (david@seclarity.io)
    Date: 06/02/2021
    This file contains functions that support capturing exactly what we need and nothing more.

    This software is provided under the Apache Software License.
    See the accompanying LICENSE file for more information.

    This file also leverages pcapy-ng, an Apache-licensed package that interfaces with the libpcap packet capture
    library. You can find more details about this repository at https://github.com/stamparm/pcapy-ng.
"""

import re
import os
import sys
import magic
import pcapy
import time
import threading
from collections import deque
from filelock import SoftFileLock, Timeout
from networksage_tools.common_utilities import packet
from networksage_tools.common_utilities import iputils
from networksage_tools.common_utilities import dnsservice
from networksage_wrappers.wrappers import wrappers as networksage

my_api_key_var = "NETWORKSAGE_API_KEY"
api_key = os.environ.get(my_api_key_var)
if api_key is None:
    print("Missing API Key. Please type export NETWORKSAGE_API_KEY='<your_api_key>' in your terminal to set up.")

def create_bpf():
    """This function creates the Berkeley Packet Filter (BPF) capture filter to restrict us from capturing (and
       therefore wasting resources on) types of traffic that are not supported by NetworkSage. The steps are outlined
       below.
    """


    """1. Create a valid BPF snippet to exclude all local protocols (those that communicate only on internal networks,
          or that have sensitive-but-uninteresting [to us] data when communicated externally).
    """
    local_protocols="445 or 5355 or netbios-ssn or netbios-ns or mdns or ldap or ldaps or bootps or bootpc or 1900"

    """2. Create a BPF snippet that confirms that at least one IP address in EVERY conversation is NOT local (still
          only works for IPv4).
    """
    local_to_local_sans_dns="(src net 0.0.0.0/32 or 192.168.0.0/16 or 10.0.0.0/8 or 172.16.0.0/12 or 239.255.255.250/32 or 224.0.0.0/4 or 255.255.255.255/32 or 127.0.0.0/8) and (dst net (0.0.0.0/32 or 192.168.0.0/16 or 10.0.0.0/8 or 172.16.0.0/12 or 239.255.255.250/32 or 224.0.0.0/4 or 255.255.255.255/32 or 127.0.0.0/8)) and not port domain"

    """3. Assemble the full BPF that (in English) does the following:
          MUST be IP and no SMB or any other E-W protocols and no non-DNS local-to-local traffic and no IPv6 (since we
          do not handle IPv6 today).
    """
    bpf_to_apply="ip and ((not port (" + local_protocols + ")) and not (" + local_to_local_sans_dns + ") and not ip6)"
    return bpf_to_apply


def capture(**kwargs):
    """This function takes an interface to capture on and a BPF capture filter and begins capturing as requested.
       Because we work with flow data collected from the IP and Transport layers, we capture a much smaller amount
       of each packet to speed up processing and stay in-memory storage efficient.
    """
    utils=kwargs['utils']
    is_verbose = kwargs['is_verbose']
    # capture just 350 bytes from each packet (some multi-answer DNS packets were being cut short @ 300 bytes)
    capture_handle=pcapy.open_live(kwargs['interface'], 350, 0, 1) # setting the timeout value to -1 is unpredictable, and 0 is waiting forever.
    capture_handle.setfilter(kwargs['bpf'])

    # get the next packet and store it in our buffer
    while not utils.stop_thread:
        (hdr, pkt) = capture_handle.next()
        while hdr is not None and not utils.stop_thread:
            utils.packet_buffer.append((hdr, pkt))
            (hdr, pkt) = capture_handle.next()
    return


def process_packets(**kwargs):
    """Processes our accumulating packet buffer in a FIFO manner. Assigns each packet to the proper Secflow.
    """
    utils = kwargs['utils']
    is_verbose = kwargs['is_verbose']
    first = True # we need to collect timestamp information on the first packet
    dns = dnsservice.DnsService(utils)  # create an instance of the DnsService class to use
    while not utils.stop_thread:
        if len(utils.packet_buffer) > 0:
            packet_data = utils.packet_buffer.popleft() # get the first (hdr, pkt) item -- we want to process in FIFO order
            hdr = packet_data[0]
            pkt = packet_data[1]
            pkt_len = len(pkt)
            if pkt_len <= 34:  # length of IPv4 packet with standard header
                print("Warning: unexpectedly short packet:\n", pkt)
            else:  # maybe legit IPv4 packet
                if first: # collect timing information
                    utils.file_start_time = min(utils.file_start_time, utils.get_timestamp_from_packet_header(hdr))
                    first=False #reset it
                packet_info = packet.PacketInfo(pkt, hdr, utils)  # prep everything for later use
                if packet_info.is_ipv4_by_ip_layer:
                    if packet_info.transport_protocol not in [6, 17, 1]: #TCP, UDP, or ICMP
                        print("Warning, expected UDP, TCP, or ICMP but got this instead:", packet_info.transport_protocol)
                        continue  # something was wrong with this packet
                else:
                    if (pkt[0] == 0 and pkt[4:6] == b"\x00\x06" and pkt[14:16] == b"\x08\x00"):
                        #this IPv4 packet has a Linux-cooked header, so ignore first 2 bytes
                        packet_info = packet.PacketInfo(pkt[2:], hdr, utils) #prep everything for later use
                    else:
                        print("Warning, either something went wrong or this is not an IPv4 packet (handling for which is unimplemented)")
                        continue # skip this session
                """Since we keep local DNS lookups in the streaming version, we need to process them here and NOT save
                   them as a Secflow.
                """
                if iputils.check_if_local_ip(packet_info.dest_ip) and iputils.check_if_local_ip(packet_info.source_ip):
                    if packet_info.is_local_dns_response(pkt):
                        dns.parse_local_lookup_from_packet(packet_info, pkt)
                        continue # we don't want to capture this as a secflow
                    else:
                        #print("Either not DNS (weird...), or a DNS lookup...skipping cataloging of this result")
                        continue
                if packet_info.protocol_information in ["ICMP"]: # handle protocols without layer 4 information specially
                    ip_a_string = packet_info.source_ip + ":" + packet_info.protocol_information
                    ip_b_string = packet_info.dest_ip + ":" + packet_info.protocol_information
                else:
                    ip_a_string = packet_info.source_ip + ":" + packet_info.source_port
                    ip_b_string = packet_info.dest_ip + ":" + packet_info.dest_port
                """Figure out if we've seen either side of the connection already. If we have, we just update the
                   existing Secflow.
                """
                if ip_a_string in utils.secflows.keys():
                    secflow = utils.secflows[ip_a_string]
                elif ip_b_string in utils.secflows.keys():
                    secflow = utils.secflows[ip_b_string]
                elif packet_info.protocol_information in ["ICMP"]:
                    if (ip_a_string + "-" + ip_b_string) in utils.secflows.keys():
                        secflow = utils.secflows[ip_a_string + "-" + ip_b_string]
                    elif ip_b_string+"-"+ip_a_string in utils.secflows.keys():
                        secflow = utils.secflows[ip_b_string + "-" + ip_a_string]
                    else:
                        secflow = packet_info.determine_session_directionality(hdr)
                else:  # we don't yet know, so go figure it out!
                    secflow = packet_info.determine_session_directionality(hdr)
                    if secflow is None:  # local-to-local DNS
                        continue # skip this session
                #print("secflow key is", secflow.key)
                if secflow.key == ip_a_string:  # we're dealing with the source
                    secflow.source_pkts += 1
                    secflow.source_payload_bytes += packet_info.upper_layer_length  # payloadLength
                    secflow.max_ts = max(secflow.max_ts, packet_info.packet_start_time)
                elif secflow.key == ip_b_string:  # we're dealing with the destination
                    secflow.dest_pkts += 1
                    secflow.dest_payload_bytes += packet_info.upper_layer_length  # payloadLength
                    secflow.max_ts = max(secflow.max_ts, packet_info.packet_start_time)
                elif packet_info.protocol_information in ["ICMP"]:
                    if secflow.key == (ip_a_string + "-" + ip_b_string):
                        secflow.source_pkts += 1
                        secflow.source_payload_bytes += packet_info.upper_layer_length  # payloadLength
                        secflow.max_ts = max(secflow.max_ts, packet_info.packet_start_time)
                    elif secflow.key == (ip_b_string + "-" + ip_a_string):
                        secflow.dest_pkts += 1
                        secflow.dest_payload_bytes += packet_info.upper_layer_length  # payloadLength
                        secflow.max_ts = max(secflow.max_ts, packet_info.packet_start_time)
                    else:
                        print("Warning, protocol without upper layer info, but seems to be deformed!")
                        continue
                else:
                    print("Warning, deformed UDP or TCP object suspected!")
                    continue
        else:
            time.sleep(1) # sleep for a second
            continue
    if utils.stop_thread:
        utils.set_secflow_durations()
        dns.get_passive_dns()
        """At this point, we want to see what existing passive DNS information was learned in recent sessions and use
           that first. TO do so, we need to lock the resource associated with the short-term pDNS file so that we don't
           make cleanup modifications to it while editing.
        """
        lock = SoftFileLock(dns.short_term_passive_dns_lock_name)
        try:
            with lock.acquire(timeout=10):
                dns.map_destination_ips_to_names()
                dns.update_passive_dns_repository() # add current file's DNS to the local PDNS file
        except Timeout:
            print("Lock acquisition for cleaning up short-term passive DNS took too long. Something might be wrong.")


def send_sample(**kwargs):
    """This function allows us to send a sample consisting of just unenriched Secflows
       to the NetworkSage uploader.
    """

    utils = kwargs['utils']
    capture_thread = kwargs['capture_thread']
    processing_thread = kwargs['processing_thread']
    is_verbose = kwargs['is_verbose']

    # stop active threads
    utils.stop_thread = True
    processing_thread.join()
    capture_thread.join(3) # we need this to terminate after a few seconds if it's stuck waiting on no packets
    uuid = utils.get_random_uuid()
    utils.set_hash_value_for_sample(uuid)

    """Capture Secflows in JSON format, and store with all information for final transmission
    """
    utils.prepare_final_output_file()

    # check if output file looks sane
    utils.check_output_sanity()

    #print("Conversion complete! Final Secflow Output stored at", utils.secflow_output_filepath)
    if api_key is None:
        print("Would be ready to send", str(len(utils.secflows)), "secflows out to NetworkSage, but no API key present.")
    else:
        with open(utils.secflow_output_filepath, 'rb') as indata:
            sample_data = indata.read()
        sample_type = "secflow"
        if len(utils.secflows) > 0:
            networksage.upload_sample(utils.sample_name, sample_data, sample_type)
            if is_verbose:
                print("Captured", str(len(utils.secflows)), "secflows.")
                print("Uploading sample "+str(uuid)+" to NetworkSage!")
                print("Data currently uploading looks like this:")
                print(utils.secflows[0])
        else:
            if is_verbose:
                print("No secflows captured. Nothing to upload.")
    utils.cleanup_files()



"""
    Copyright (c) 2021 David Pearson (david@seclarity.io)
    Date: 06/02/2021
    This file contains utilities that make processing a .cap, .pcap, or .pcapng file easy to do. Ultimately, the final
    function in this file will create secflows out of the incoming capture data.

    This software is provided under the Apache Software License.
    See the accompanying LICENSE file for more information.

    This file also leverages pcapy-ng, an Apache-licensed package that interfaces with the libpcap packet capture
    library. You can find more details about this repository at https://github.com/stamparm/pcapy-ng.
"""

import re
import sys
import magic
import pcapy
from networksage_tools.common_utilities import packet

def filter_capture_file_by_bpf(utils, bpf_to_apply):
    """Takes a Berkeley packet filter (BPF) and applies it to the capture file. Whatever remains after the BPF is
       applied will be saved to the output file.
    """

    last_dot = utils.original_filepath.rfind(".")
    if last_dot == -1:  # no extension in file name
        utils.filtered_filepath = utils.original_filepath + "_filtered"
    else:
        utils.filtered_filepath = utils.original_filepath[:last_dot] + "_filtered" + utils.original_filepath[last_dot:]

    capfile = pcapy.open_offline(utils.original_filepath)
    output = capfile.dump_open(utils.filtered_filepath)
    capfile.setfilter(bpf_to_apply)
    (hdr, pkt) = capfile.next()

    while hdr is not None:
        output.dump(hdr, pkt)
        (hdr, pkt) = capfile.next()
    del output


def remove_local_traffic_from_capture_file(utils, dns):
    """Calls most of the functions to prepare and then remove the local traffic from the incoming capture file.
    """

    bpf_to_apply = ""  # set up the empty BPF

    """Remove all of the local/uninteresting traffic
         Steps:

         1. Get all of the DNS queries in form <port> <dns.qry.name> <dns.resp.name>
    """
    dns.parse_dns_records_from_capture_file()

    """2. Identify which of the DNS queries are either local forward or local reverse lookups (looking for
          things in RFC1918 and similar IP address ranges). Note that this currently only works for IPv4.
    """
    local_lookups = dns.collect_local_lookups()

    """3. Create a valid Berkeley Packet Filter (BPF) to exclude all local protocols (those that communicate only on
          internal networks, or that have sensitive-but-uninteresting [to us] data when communicated externally).
    """
    local_protocols = "445 or 5355 or netbios-ssn or netbios-ns or mdns or ldap or ldaps or bootps or bootpc or 1900"

    """4. Create a BPF that confirms that at least one IP address in EVERY conversation is NOT local (still only works
          for IPv4).
    """
    local_to_local_sans_dns = "(src net 0.0.0.0/32 or 192.168.0.0/16 or 10.0.0.0/8 or 172.16.0.0/12 or 239.255.255.250/32 or 224.0.0.0/4 or 255.255.255.255/32 or 127.0.0.0/8) and (dst net (0.0.0.0/32 or 192.168.0.0/16 or 10.0.0.0/8 or 172.16.0.0/12 or 239.255.255.250/32 or 224.0.0.0/4 or 255.255.255.255/32 or 127.0.0.0/8)) and not port domain"

    """5. Create a BPF that excludes DNS resolutions using internal resolvers, and explicitly excludes those
          local lookups (by destination port) that we found earlier.
    """
    dns_prefix = " and not (port domain and (src net 10.0.0.0/8 or 192.168.0.0/16 or 172.16.0.0/12) and (dst net 10.0.0.0/8 or 192.168.0.0/16 or 172.16.0.0/12)"

    dns_expression = ""
    if len(local_lookups) > 0:
        dns_expression += (dns_prefix + " and port (" + "".join(local_lookups) + "))")

    """6. Assemble the full BPF that (in English) does the following:
            MUST be IP and no SMB or any other E-W protocols and no non-DNS local-to-local traffic and no local-to-local
            DNS traffic with local lookups and no IPv6 (since we do not handle IPv6 today).
    """
    bpf_to_apply = "ip and ((not port (" + local_protocols + ")) and not (" + local_to_local_sans_dns + ") " + dns_expression + " and not ip6)"

    """7. Filter the capture file using the assembled BPF, and then return the filtered file location for use in the
          rest of the pipeline.
    """
    filter_capture_file_by_bpf(utils, bpf_to_apply)


def validate_file_format(utils):
    """Validate that the file is of an accepted type (CAP, PCAP, or PCAPNG).
    """

    # check if file is valid
    if not re.match(r"^(p|)cap(|(|\-)ng) capture file", magic.from_file(utils.original_filepath)):
        print("Error:", utils.original_filepath + ",", "of type", magic.from_file(utils.original_filepath),
              "is not an accepted file type.")
        sys.exit(1)


def pcap_to_secflow_converter(utils, dns):
    """Given a capture file that has been validated, remove any local traffic and collect
        the secflows for all TCP, UDP, and ICMP sessions.
    """
    # filter the original file to remove local-to-local traffic
    remove_local_traffic_from_capture_file(utils, dns)

    get_min_timestamp_from_capture_file(utils)

    # get all of the UDP records in place
    parse_secflows_from_capture_file(utils, "udp")

    # get all of the TCP records in place
    parse_secflows_from_capture_file(utils, "tcp")

    # get all of the ICMP records in place
    parse_secflows_from_capture_file(utils, "icmp")


def get_min_timestamp_from_capture_file(utils):
    """Iterates through the file to find the absolute minimum timestamp.
    """
    with pcapy.open_offline(utils.filtered_filepath) as capfile:
        capfile.setfilter("udp or tcp or icmp")
        (hdr, pkt) = capfile.next()
        while hdr is not None:
            # update start time with the actual file min
            utils.file_start_time = min(utils.file_start_time, utils.get_timestamp_from_packet_header(hdr))
            (hdr, pkt) = capfile.next()


def parse_secflows_from_capture_file(utils, protocol):
    """Takes a capture file and a transport protocol, parses all relevant packets to find the directionality of the
       sessions to which they belong, and then creates and returns an ordered dictionary of secflows.
    """
    with pcapy.open_offline(utils.filtered_filepath) as capfile:
        if not re.match("^(tc|ud|icm)p$", protocol):
            print("Error, protocol", protocol, "not supported!")
            sys.exit(1)
        capfile.setfilter(protocol)
        (hdr, pkt) = capfile.next()
        while hdr is not None:
            # we should only be getting TCP, UDP, or ICMP records now, so parse accordingly
            pkt_len = len(pkt)
            if pkt_len <= 34:  # length of IPv4 packet with standard header
                print("Warning: unexpectedly short packet:\n", pkt)
            else:  # maybe legit IPv4 packet
                packet_info = packet.PacketInfo(pkt, hdr, utils)  # prep everything for later use
                if packet_info.is_ipv4_by_ip_layer:
                    if packet_info.transport_protocol not in [6, 17, 1]: # TCP, UDP, or ICMP
                        print("Warning, expected UDP, TCP, or ICMP but got this instead:", packet_info.transport_protocol)
                        continue  # something was wrong with this packet
                else:
                    if (pkt[0] == 0 and pkt[4:6] == b"\x00\x06" and pkt[14:16] == b"\x08\x00"):
                        #this IPv4 packet has a Linux-cooked header, so ignore first 2 bytes
                        packet_info = packet.PacketInfo(pkt[2:], hdr, utils) #prep everything for later use
                    else:
                        print("Warning, either something went wrong or this is not an IPv4 packet (handling for which is unimplemented)")
                        (hdr, pkt) = capfile.next() #skip this session
                        continue

                if packet_info.protocol_information in ["ICMP"]: # handle protocols without layer 4 information specially
                    ip_a_string = packet_info.source_ip + ":" + packet_info.protocol_information
                    ip_b_string = packet_info.dest_ip + ":" + packet_info.protocol_information
                else:
                    ip_a_string = packet_info.source_ip + ":" + packet_info.source_port
                    ip_b_string = packet_info.dest_ip + ":" + packet_info.dest_port
                #print("Checking", ip_a_string, "and", ip_b_string)

                """ Figure out if we've seen either side of the connection already. If we have, we just update the existing secflow.
                """
                if ip_a_string in utils.secflows.keys():
                    secflow = utils.secflows[ip_a_string]
                elif ip_b_string in utils.secflows.keys():
                    secflow = utils.secflows[ip_b_string]
                elif packet_info.protocol_information in ["ICMP"]:
                    if (ip_a_string + "-" + ip_b_string) in utils.secflows.keys():
                        secflow = utils.secflows[ip_a_string + "-" + ip_b_string]
                    elif ip_b_string + "-" + ip_a_string in utils.secflows.keys():
                        secflow = utils.secflows[ip_b_string + "-" + ip_a_string]
                    else:
                        secflow = packet_info.determine_session_directionality(hdr)
                else:  # we don't yet know, so go figure it out!
                    secflow = packet_info.determine_session_directionality(hdr)
                    if secflow is None:  # local-to-local DNS
                        (hdr, pkt) = capfile.next()  # skip this session
                        continue
                if secflow.key == ip_a_string:  # we're dealing with the source
                    secflow.source_pkts += 1
                    secflow.source_payload_bytes += packet_info.upper_layer_length  # payloadLength
                    secflow.max_ts = max(secflow.max_ts, packet_info.packet_start_time)
                elif secflow.key == ip_b_string:  # we're dealing with the destination
                    secflow.dest_pkts += 1
                    secflow.dest_payload_bytes += packet_info.upper_layer_length  # payloadLength
                    secflow.max_ts = max(secflow.max_ts, packet_info.packet_start_time)
                elif packet_info.protocol_information in ["ICMP"]:
                    if secflow.key == (ip_a_string + "-" + ip_b_string):
                        secflow.source_pkts += 1
                        secflow.source_payload_bytes += packet_info.upper_layer_length  # payloadLength
                        secflow.max_ts = max(secflow.max_ts, packet_info.packet_start_time)
                    elif secflow.key == (ip_b_string + "-" + ip_a_string):
                        secflow.dest_pkts += 1
                        secflow.dest_payload_bytes += packet_info.upper_layer_length  # payloadLength
                        secflow.max_ts = max(secflow.max_ts, packet_info.packet_start_time)
                    else:
                        print("Warning, protocol without upper layer info, but seems to be deformed!")
                        continue
                else:
                    print("Warning, deformed UDP or TCP object suspected!")
                    continue
            (hdr, pkt) = capfile.next()



"""
    Copyright (c) 2021 David Pearson (david@seclarity.io)
    Date: 06/03/2021
    This file contains utility functions and variables that are used by a number of callers.

    This software is provided under the Apache Software License.
    See the accompanying LICENSE file for more information.
"""

import collections
import json
import os
import sys
import uuid
from pathlib import Path

class Utilities():
    local_ips_forward_regex = r"^(10|127|169\.254|172\.1[6-9]|172\.2[0-9]|172\.3[0-1]|192\.168)\."
    local_ips_reverse_regex = r"([0-9]{1,3})\.([0-9]{1,3})\.((1[6-9]\.172)|(2[0-9]\.172)|(3[0-1]\.172)|([0-9]{1,3}\.10)|(168\.192))\.in\-addr\.arpa\.$"
    ipv6_regex = r"(([1-9a-fA-F]{1,}:)"  # very basic, meant to filter out virtually anything right now. Also not currently used

    def __init__(self, orig_fp, platform_type, output_dir=None):
        self.active_external_ips = set()
        self.questions = dict()
        self.original_filepath = orig_fp
        if orig_fp is None:
            self.is_streaming = True
        else:
            self.is_streaming = False
        self.output_dir = output_dir # optional place to store the finished file
        self.system_type = platform_type
        self.filtered_filepath = ""
        self.secflows = collections.OrderedDict()  # dict of Secflow objects (TCP, UDP, or ICMP)
        self.file_start_time = float(99999999999)  # set it impossibly high at beginning
        self.tmp_filepath = ""  # file to temporarily store just the secflows with proper formatting. Needed in order to have a consistent file to hash
        self.secflow_output_filepath = ""  # file to store the final output intended to be sent back to the caller
        self.sample_name = ""
        self.secflows_hash = ""  # used to store the hash of the secflows file
        self.zeekflows = collections.OrderedDict()
        self.packet_buffer = collections.deque() #create an empty buffer to store packets that have not yet been processed
        self.stop_thread = False #keep state of whether thread should be stopped

    def cleanup_files(self):
        """Removes any files that were created in the interim but not needed for the final output.
        """
        if not self.zeekflows:  # we were working with PCAP data
            # delete the filtered file, if it exists
            try:
                os.remove(self.filtered_filepath)
            except OSError:
                pass
        # delete the temporary file, if it exists
        try:
            os.remove(self.tmp_filepath)
        except OSError:
            pass
        if self.is_streaming or len(self.secflows) == 0: # only delete if streaming or empty
            # delete the final output .sf file, if it exists
            try:
                os.remove(self.secflow_output_filepath)
            except OSError:
                pass

    def get_timestamp_from_packet_header(self, packet_header):
        """Simple utility to transform packet_header's startTime into a usable float value.
        """
        return float(str(packet_header.getts()[0]) + "." + str(packet_header.getts()[1]).zfill(6))

    def set_secflow_durations(self):
        """Iterates through all secflows and updates their flow duration fields.
        """
        for secflow in self.secflows:
            self.secflows[secflow].set_secflow_duration()

    def save_secflows_to_file(self):
        """Saves secflows to a .sf file for hashing.
        """
        # get the proper file name
        last_dot = self.filtered_filepath.rfind(".")
        if last_dot == -1:  # no extension in file name
            self.tmp_filepath = self.filtered_filepath + ".tmp"
        else:
            self.tmp_filepath = self.filtered_filepath[:last_dot] + ".tmp"

        with open(self.tmp_filepath, "w") as sf_out:
            for secflow in self.secflows.keys():
                if self.secflows[secflow].protocol_information in ["ICMP"]:
                    first_delimiter = self.secflows[secflow].key.find(":")
                    sf_out.write(self.secflows[secflow].key[:first_delimiter] + " <-> " +
                                self.secflows[secflow].dest_name + "\t"
                                + str(self.secflows[secflow].source_pkts) + "\t"
                                + str(self.secflows[secflow].source_payload_bytes) + "\t"
                                + str(self.secflows[secflow].dest_pkts) + "\t"
                                + str(self.secflows[secflow].dest_payload_bytes) + "\t"
                                + str(self.secflows[secflow].relative_start_time) + "\t"
                                + str(self.secflows[secflow].protocol_information) + "\t"
                                + str(self.secflows[secflow].duration) + "\n")
                else:
                    sf_out.write(self.secflows[secflow].key + " <-> " +
                                self.secflows[secflow].dest_name + ":" + self.secflows[secflow].dest_port + "\t"
                                + str(self.secflows[secflow].source_pkts) + "\t"
                                + str(self.secflows[secflow].source_payload_bytes) + "\t"
                                + str(self.secflows[secflow].dest_pkts) + "\t"
                                + str(self.secflows[secflow].dest_payload_bytes) + "\t"
                                + str(self.secflows[secflow].relative_start_time) + "\t"
                                + str(self.secflows[secflow].protocol_information) + "\t"
                                + str(self.secflows[secflow].duration) + "\n")

    def get_random_uuid(self) -> object:
        """Assigns a random uuid to the sample file
        """
        return str(uuid.uuid4()).replace("-", "")

    def set_hash_value_for_sample(self, uuid) -> object:
        """Assigns a uuid to the sample file
        """
        self.secflows_hash = uuid

    def prepare_final_output_file(self):
        """Captures Secflows and other important metadata in JSON format for final transmission back to the caller
            (usually an API).
        """
        if self.is_streaming: # this is streaming mode
            time_str = str(self.file_start_time)
            start_time_str = str(time_str[:time_str.find(".")])
            self.secflow_output_filepath = self.system_type + "_" + start_time_str + ".sf"
        else:
            if not self.is_streaming and not Path(self.tmp_filepath).is_file():
                print("Warning:", self.tmp_filepath, "does not exist. Aborting.")
                sys.exit(1)
            else:
                # get the proper file name
                last_dot = self.filtered_filepath.rfind(".")
                if last_dot == -1:  # no extension in file name
                    self.secflow_output_filepath = self.filtered_filepath + ".sf"
                else:
                    self.secflow_output_filepath = self.filtered_filepath[:last_dot] + ".sf"

        """Use the dictionary version of this file to convert it (plus some other information) to JSON
        """
        # first, make sure we're in ascending relativeStart order
        self.secflows = {k: v for k, v in sorted(self.secflows.items(), key=lambda item: item[1].relative_start_time)}

        with open(self.secflow_output_filepath, "w") as secflow_outfile:
            self.sample_name = os.path.basename(self.secflow_output_filepath)
            rows = []
            for flow in self.secflows.keys():
                if self.secflows[flow].protocol_information in ["ICMP"]:
                    first_delimiter = self.secflows[flow].key.find(":")
                    rows += [{"src": self.secflows[flow].key[:first_delimiter],
                              "dst": self.secflows[flow].dest_name,
                              "destinationNameSource": self.secflows[flow].destination_name_source,
                              "srcPkts": self.secflows[flow].source_pkts,
                              "srcBytes": self.secflows[flow].source_payload_bytes,
                              "dstPkts": self.secflows[flow].dest_pkts, "dstBytes": self.secflows[flow].dest_payload_bytes,
                              "relativeStart": self.secflows[flow].relative_start_time,
                              "protocolInformation": self.secflows[flow].protocol_information,
                              "duration": self.secflows[flow].duration}]
                else:
                    rows += [{"src": self.secflows[flow].key,
                              "dst": self.secflows[flow].dest_name + ":" + self.secflows[flow].dest_port,
                              "destinationNameSource": self.secflows[flow].destination_name_source,
                              "srcPkts": self.secflows[flow].source_pkts,
                              "srcBytes": self.secflows[flow].source_payload_bytes,
                              "dstPkts": self.secflows[flow].dest_pkts, "dstBytes": self.secflows[flow].dest_payload_bytes,
                              "relativeStart": self.secflows[flow].relative_start_time,
                              "protocolInformation": self.secflows[flow].protocol_information,
                              "duration": self.secflows[flow].duration}]
            # metadata for the file that is important to keep track of
            secflow_output = {"hash": self.secflows_hash, "trafficDate": str(self.file_start_time),
                          "fileName": self.sample_name,
                          "flashes": rows}  # note: can rename flashes to secflows
            json.dump(secflow_output, secflow_outfile)
            return self.secflow_output_filepath

    def check_output_sanity(self):
        """Determines if there are any issues with the final output file. Issues could include:
                1. having no resulting FLASHES (this is atypical but not an error in this case)
                2. having a default time (meaning that we didn't process anything)
        """

        try:
            with open(self.secflow_output_filepath, "r") as secflow_outfile:
                out_file = json.load(secflow_outfile)
                if out_file["trafficDate"] == "99999999999.0": #default date value
                    if len(out_file["flashes"]) == 0:
                        print("File has no secflows to upload. NetworkSage currently handles only IPv4 to or from the Internet. Nothing to do.")
                        return
                    else:
                        print("Error: File seems to have secflows to upload, but date of traffic wasn't correctly learned. Aborting.")
                        sys.exit(1)
        except:
            print("Something went wrong while trying to check output file for sanity. Aborting.")
            self.cleanup_files()
            sys.exit(1)